(* large cardinals, solyga@gmx.de, 2022-09-14 *)
IMPLEMENTATION MODULE LC;

FROM SYSTEM IMPORT BYTE;
IMPORT Types, Lib, Lib2;
IMPORT IO;

CONST
  baseMax = 10 + ( ORD('Z') + 1 - ORD('A') );

TYPE
  TCard = Types.TCard;
  TInt = Types.TInt;


PROCEDURE WrHex( x-: ARRAY OF BYTE );
VAR
  s: ARRAY [0..127] OF CHAR;
  ok: BOOLEAN;
BEGIN
  CardToStr( x, s, 16, ok );
  IF ok THEN
    IO.WrStr( s );
  ELSE
    IO.WrStr( '???' );
  END;
END WrHex;

CONST
  WrLn ::= IO.WrLn;
  WrStr ::= IO.WrStr;
  WrCard ::= IO.WrCard;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Lib.FatalError( s );
  HALT;
END Fatal;


PROCEDURE Char( x: Types.TShtCard ): CHAR;
BEGIN
  IF x < 10 THEN
    RETURN CHR( x + ORD('0') );
  ELSIF x < baseMax THEN
    RETURN CHR( x - 10 + ORD('A') );
  ELSE
    RETURN '?';  (* ??? *)
  END;
END Char;

PROCEDURE Invert( VAR s: ARRAY OF CHAR; l: CARDINAL );
VAR
  i, j: CARDINAL;
  c: CHAR;
BEGIN
  IF l > SIZE( s ) THEN l := SIZE( s ) END;
  IF l < 2 THEN RETURN END;
  FOR i := 0 TO l DIV 2 - 1 DO
    j := l - 1 - i;
    c := s[ i ]; s[ i ] := s[ j ]; s[ j ] := c;
  END;
END Invert;

PROCEDURE CardToStr( x-: ARRAY OF BYTE; VAR s: ARRAY OF CHAR; b: CARDINAL; VAR ok: BOOLEAN );
VAR
  l: CARDINAL;

  PROCEDURE Put( c: CHAR );
  BEGIN
    IF l >= SIZE( s ) THEN ok := FALSE; RETURN END;
    s[ l ] := c;
    INC( l );
  END Put;

VAR
  i: CARDINAL;
  y: Types.TShtCard;
BEGIN
  l := 0;
  ok := TRUE;
  IF b # 16 THEN ok := FALSE; RETURN END;
  FOR i := 0 TO SIZE( x ) - 1 DO
    y := x[ i ];
    Put( Char( y MOD 10H ) );
    Put( Char( y DIV 10H ) );
  END;
  WHILE ( l > 1 ) & ( s[l-1] = '0' ) DO DEC( l ) END;
  IF l < SIZE( s ) THEN s[ l ] := 0C END;
  Invert( s, l );
END CardToStr;


<* PUSH *> <* CHECKINDEX - *>
PROCEDURE Len( x-: ARRAY OF BYTE ): CARDINAL;
VAR
  l: CARDINAL;
BEGIN
  l := SIZE( x );
  WHILE ( l > 1 ) & ( x[l-1] = BYTE(0) ) DO DEC( l ) END;
  RETURN l;
END Len;

PROCEDURE Len0( x-: ARRAY OF BYTE ): CARDINAL;
VAR
  l: CARDINAL;
BEGIN
  l := SIZE( x );
  WHILE ( l > 0 ) & ( x[l-1] = BYTE(0) ) DO DEC( l ) END;
  RETURN l;
END Len0;

PROCEDURE LenI( x-: ARRAY OF BYTE; i: CARDINAL ): CARDINAL;
VAR
  l: CARDINAL;
BEGIN
  l := SIZE( x );
  IF l <= i THEN RETURN 0 END;
  WHILE ( l > i ) & ( x[l-1] = BYTE(0) ) DO DEC( l ) END;
  RETURN l - i;
END LenI;
<* POP *>


<* PUSH *> <* CHECKINDEX - *>
PROCEDURE Cpy( x-: ARRAY OF BYTE; VAR y: ARRAY OF BYTE ): BOOLEAN;
VAR
  l, i: CARDINAL;
BEGIN
  l := Len( x );
  IF l > SIZE( y ) THEN RETURN FALSE END;
  FOR i := 0 TO l - 1 DO y[i] := x[i] END;
  WHILE l < SIZE( y ) DO y[l] := BYTE(0); INC( l ) END;
  RETURN TRUE;
END Cpy;

PROCEDURE Cmp( x-, y-: ARRAY OF BYTE ): Types.TShtInt;
VAR
  lx, ly: CARDINAL;
  cx, cy: TSht;
BEGIN
  lx := Len( x ); ly := Len( y );
  IF lx < ly THEN RETURN -1 ELSIF lx > ly THEN RETURN 1 END;
  WHILE lx > 0 DO
    DEC( lx );
    cx := x[lx]; cy := y[lx];
    IF cx < cy THEN RETURN -1 ELSIF cx > cy THEN RETURN 1 END;
  END;
  RETURN 0;
END Cmp;
<* POP *>


<* PUSH *> <* CHECKINDEX - *>
PROCEDURE Inc( VAR x: ARRAY OF BYTE; y-: ARRAY OF BYTE ): BOOLEAN;
TYPE
  TCnvDesc = RECORD
    CASE : CARDINAL OF
    | 0: v: TCard;
    | 1: l, h: TSht;
    END;
  END;
VAR
  i, l: CARDINAL;
  c: TCnvDesc;
BEGIN
  i := 0; l := Len( y ); c.h := 0;
  WHILE i < l DO
    IF i >= SIZE( x ) THEN RETURN FALSE END;
    c.v := TSht( x[i] ) + TSht( y[i] ) + c.h;
    x[i] := c.l;
    INC( i );
  END;
  WHILE c.h > 0 DO
    IF i >= SIZE( x ) THEN RETURN FALSE END;
    c.v := TSht( x[i] ) + c.h;
    x[i] := c.l;
    INC( i );
  END;
  RETURN TRUE;
END Inc;

PROCEDURE Add( x-, y-: ARRAY OF BYTE; VAR z: ARRAY OF BYTE ): BOOLEAN;
BEGIN
  RETURN Cpy( x, z ) & Inc( z, y );
END Add;
<* POP *>


<* PUSH *> <* CHECKINDEX - *>
PROCEDURE Dec( VAR x: ARRAY OF BYTE; y-: ARRAY OF BYTE ): BOOLEAN;
TYPE
  TCnvDesc = RECORD
    CASE : CARDINAL OF
    | 0: v: TInt;
    | 1: l, h: TSht;
    END;
  END;
VAR
  i, l: CARDINAL;
  c: TCnvDesc;
BEGIN
  i := 0; l := Len0( y ); c.h := 0;
  WHILE i < l DO
    IF i >= SIZE( x ) THEN RETURN FALSE END;
    c.v := TInt( TSht( x[i] ) ) - TSht( y[i] ) - c.h;
    IF c.v < 0 THEN c.h := 1 END;
    x[i] := c.l;
    INC( i );
  END;
  WHILE c.h > 0 DO
    IF i >= SIZE( x ) THEN RETURN FALSE END;
    c.v := TInt( TSht( x[i] ) ) - c.h;
    IF c.v < 0 THEN c.h := 1 END;
    x[i] := c.l;
    INC( i );
  END;
  RETURN TRUE;
END Dec;

PROCEDURE Sub( x, y-: ARRAY OF BYTE; VAR z: ARRAY OF BYTE ): BOOLEAN;
BEGIN
  RETURN Dec( x, y ) & Cpy( x, z );
END Sub;
<* POP *>


<* PUSH *> <* CHECKINDEX - *>
PROCEDURE Mul( x-, y-: ARRAY OF BYTE; VAR z: ARRAY OF BYTE ): BOOLEAN;
TYPE
  TCnvDesc = RECORD
    CASE : CARDINAL OF
    | 0: v: TCard;
    | 1: l, h: TSht;
    END;
  END;
VAR
  ix, lx, iy, ly, i: CARDINAL;
  f: TSht;
  c: TCnvDesc;
BEGIN
  lx := Len( x ); ly := Len( y );
  IF lx + ly - 1 > SIZE( z ) THEN RETURN FALSE END;
  Lib2.Fill( z, 0 );
  FOR iy := 0 TO ly - 1 DO
    f := TSht( y[iy] );
    c.h := 0;
    FOR ix := 0 TO lx - 1 DO
      i := ix + iy;
      c.v := TSht( z[i] ) + TSht( x[ix] ) * f + c.h;  (* FF + FF * FF + FF = FFFF *)
      z[i] := c.l;
    END;
    IF c.h > 0 THEN
      i := lx + iy;
      IF i >= SIZE( z ) THEN RETURN FALSE END;
      z[i] := c.h;
    END;
  END;
  RETURN TRUE;
END Mul;
<* POP *>


<* PUSH *> <* CHECKINDEX - *>
PROCEDURE Div( x, y-: ARRAY OF BYTE; VAR z, r: ARRAY OF BYTE ): BOOLEAN;

  PROCEDURE Cmp( x-: ARRAY OF BYTE; ix: CARDINAL; y-: ARRAY OF BYTE ): INTEGER;
  VAR
    lx, ly: CARDINAL;
    cx, cy: TSht;
  BEGIN
    lx := LenI( x, ix ); ly := Len0( y );
    IF lx < ly THEN RETURN -1 ELSIF lx > ly THEN RETURN 1 END;
    WHILE lx > 0 DO
      DEC( lx );
      cx := x[ix+lx]; cy := y[lx];
      IF cx < cy THEN RETURN -1 ELSIF cx > cy THEN RETURN 1 END;
    END;
    RETURN 0;
  END Cmp;

  PROCEDURE Dec( VAR x: ARRAY OF BYTE; ix: CARDINAL; y-: ARRAY OF BYTE );
  TYPE
    TCnvDesc = RECORD
      CASE : CARDINAL OF
      | 0: v: TInt;
      | 1: l, h: TSht;
      END;
    END;
  VAR
    iy, ly: CARDINAL;
    c: TCnvDesc;
  BEGIN
    iy := 0; ly := Len0( y ); c.h := 0;
    WHILE iy < ly DO
      c.v := TInt( TSht( x[ix] ) ) - TSht( y[iy] ) - c.h;
      IF c.v < 0 THEN c.h := 1 END;
      x[ix] := c.l;
      INC( ix ); INC( iy );
    END;
    IF c.h > 0 THEN x[ix] := BYTE( TSht( x[ix] ) - 1 ) END;
  END Dec;

  PROCEDURE Inc( VAR x: BYTE );
  BEGIN
    x := BYTE( TSht(x) + 1 );
  END Inc;

VAR
  l: CARDINAL;
BEGIN
  IF Len0( y ) = 0 THEN RETURN FALSE END;
  Lib2.Fill( z, 0 );
  l := Len( x );
  WHILE l > 0 DO
    DEC( l );
    WHILE Cmp( x, l, y ) >= 0 DO
      Dec( x, l, y );
      Inc( z[l] );
    END;
  END;
  RETURN Cpy( x, r );
END Div;
<* POP *>

END LC.
