MODULE LrgCards;

FROM SYSTEM IMPORT BYTE;
IMPORT Types, LC, Config;
IMPORT IO, Lib;
IMPORT WinBase, Tmr0;


CONST
  Now ::= WinBase.GetTickCount;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  IO.WrLn();
  Lib.FatalError( s );
  HALT;
END Fatal;


PROCEDURE WrHex( x-: ARRAY OF BYTE );
VAR
  s: ARRAY [0..127] OF CHAR;
  ok: BOOLEAN;
BEGIN
  LC.CardToStr( x, s, 16, ok );
  IF ok THEN
    IO.WrStr( s );
  ELSE
    IO.WrStr( '???' );
  END;
END WrHex;

CONST
  WrLn ::= IO.WrLn;
  WrStr ::= IO.WrStr;


PROCEDURE Test00();
VAR i, j: CARDINAL;
BEGIN
  FOR i := 0 TO 2 DO
    IO.WrCard( i, 1 ); IO.WrLn();
  END;
  j := i;
  IO.WrLngCard( j, 1 ); IO.WrLn();
END Test00;

PROCEDURE Test01();

  PROCEDURE Use( x-: ARRAY OF BYTE; i: CARDINAL );
  BEGIN
    <* PUSH *> <* CHECKINDEX - *>
    IO.WrHex( ORD( x[i] ), 3 ); IO.WrLn();
    <* POP *>
  END Use;

VAR
  x: ARRAY [0..1] OF CHAR;
  i: CARDINAL;
BEGIN
  FOR i := 0 TO 2 DO Use( x, i ) END;;
END Test01;


PROCEDURE Test10();
VAR
  x: Types.TShtCard;
  y: Types.TCard;
  t: Tmr0.TTicks;
BEGIN
  IO.WrStr( 'Test10: begin' ); IO.WrLn();
  t := Now();

  x := 0CFH;
  y := 11H;
  WrStr( '  INC( ' ); WrHex( x ); WrStr( ', ' ); WrHex( y ); WrStr( ' ) = ' );
  IF ~ LC.Inc( x, y ) THEN Fatal( 'Test10: Inc failed' ) END;
  WrHex( x ); WrLn();

  IO.WrStr( 'Test10: end ' ); IO.WrLngCard( Tmr0.Diff( t, Now() ), 1 ); IO.WrStr( 'ms' ); IO.WrLn();
END Test10;

PROCEDURE Test11();
VAR
  x: Types.TLngCard;
  y: Types.TLngCard;
  z: LC.TCard05;
  t: Tmr0.TTicks;
BEGIN
  IO.WrStr( 'Test11: begin' ); IO.WrLn();
  t := Now();

  x := 0FFFFEFEFH;
  y := 01111H;
  IF ~ LC.Add( x, y, z ) THEN Fatal( 'Test11: Add failed' ) END;
  WrStr( '  ' ); WrHex( x ); WrStr( ' + ' ); WrHex( y ); WrStr( ' = ' ); WrHex( z ); WrLn();

  IO.WrStr( 'Test11: end ' ); IO.WrLngCard( Tmr0.Diff( t, Now() ), 1 ); IO.WrStr( 'ms' ); IO.WrLn();
END Test11;

PROCEDURE Test20();
VAR
  x, y: Types.TCard;
  t: Tmr0.TTicks;
BEGIN
  IO.WrStr( 'Test20: begin' ); IO.WrLn();
  t := Now();

  x := 0FFFEH;
  y := 0FFH;
  WrStr( '  DEC( ' ); WrHex( x ); WrStr( ', ' ); WrHex( y ); WrStr( ' ) = ' );
  IF ~ LC.Dec( x, y ) THEN Fatal( 'Test20: Dec failed' ) END;
  WrHex( x ); WrLn();

  IO.WrStr( 'Test20: end ' ); IO.WrLngCard( Tmr0.Diff( t, Now() ), 1 ); IO.WrStr( 'ms' ); IO.WrLn();
END Test20;

PROCEDURE Test21();
VAR
  x: Types.TCard;
  y: Types.TShtCard;
  z: LC.TCard01;
  t: Tmr0.TTicks;
BEGIN
  IO.WrStr( 'Test21: begin' ); IO.WrLn();
  t := Now();

  x := 0100H;
  y := 01H;
  IF ~ LC.Sub( x, y, z ) THEN Fatal( 'Test21: Sub failed' ) END;
  WrStr( '  ' ); WrHex( x ); WrStr( ' - ' ); WrHex( y ); WrStr( ' = ' ); WrHex( z ); WrLn();

  IO.WrStr( 'Test21: end ' ); IO.WrLngCard( Tmr0.Diff( t, Now() ), 1 ); IO.WrStr( 'ms' ); IO.WrLn();
END Test21;

PROCEDURE Test30();
VAR
  x: Types.TCard;
  y: Types.TCard;
  z: LC.TCard03;
  t: Tmr0.TTicks;
BEGIN
  IO.WrStr( 'Test30: begin' ); IO.WrLn();
  t := Now();

  x := 0FFFFH;
  y := 0FFH;
  IF ~ LC.Mul( x, y, z ) THEN Fatal( 'Test30: Mul failed' ) END;
  WrStr( '  ' ); WrHex( x ); WrStr( ' * ' ); WrHex( y ); WrStr( ' = ' ); WrHex( z ); WrLn();

  IO.WrStr( 'Test30: end ' ); IO.WrLngCard( Tmr0.Diff( t, Now() ), 1 ); IO.WrStr( 'ms' ); IO.WrLn();
END Test30;

PROCEDURE Test31();
VAR
  x: Types.TCard;
  y: Types.TCard;
  z: Types.TLngCard;
  t: Tmr0.TTicks;
BEGIN
  IO.WrStr( 'Test31: begin' ); IO.WrLn();
  t := Now();

  FOR x := 0 TO 0FFFFH DO
    IF x MOD 400H = 0 THEN WrHex( x ); WrLn(); END;
    FOR y := 0 TO 0FFFFH DO
      IF ~ LC.Mul( x, y, z ) THEN Fatal( 'Test31: Mul failed' ) END;
      IF z # x * y THEN
        WrStr( ' x = ' ); WrHex( x ); WrLn();
        WrStr( ' y = ' ); WrHex( y ); WrLn();
        WrStr( ' z = ' ); WrHex( z ); WrLn();
        Fatal( 'Test31: Wrong result' );
      END;
    END;
  END;

  IO.WrStr( 'Test31: end ' ); IO.WrLngCard( Tmr0.Diff( t, Now() ), 1 ); IO.WrStr( 'ms' ); IO.WrLn();
END Test31;

PROCEDURE Test40();
VAR
  x: Types.TLngCard;
  y: Types.TShtCard;
  z: LC.TCard03;
  r: LC.TCard02;
  t: Tmr0.TTicks;
BEGIN
  IO.WrStr( 'Test40: begin' ); IO.WrLn();
  t := Now();

  x := 0FF0000H;
  y := 0FFH;
  IF ~ LC.Div( x, y, z, r ) THEN Fatal( 'Test40: Div failed' ) END;
  WrStr( '  ' ); WrHex( x ); WrStr( ' DIV ' ); WrHex( y ); WrStr( ' = ' ); WrHex( z ); WrLn();
  WrStr( '  ' ); WrHex( x ); WrStr( ' MOD ' ); WrHex( y ); WrStr( ' = ' ); WrHex( r ); WrLn();

  IO.WrStr( 'Test40: end ' ); IO.WrLngCard( Tmr0.Diff( t, Now() ), 1 ); IO.WrStr( 'ms' ); IO.WrLn();
END Test40;

PROCEDURE Test41();
VAR
  x: Types.TLngCard;
  y: Types.TCard;
  z: Types.TLngCard;
  r: Types.TCard;
  t: Tmr0.TTicks;
BEGIN
  IO.WrStr( 'Test41: begin' ); IO.WrLn();
  t := Now();

  FOR x := 0 TO 0FFFFFFFFH DO
    IF x MOD 400H = 0 THEN WrHex( x ); WrLn(); END;
    FOR y := 1 TO 0FFFFH DO
      IF ~ LC.Div( x, y, z, r ) THEN Fatal( 'Test41: Div failed' ) END;
      IF ( z # x DIV y ) OR ( r # x MOD y ) THEN
        WrStr( ' x = ' ); WrHex( x ); WrLn();
        WrStr( ' y = ' ); WrHex( y ); WrLn();
        WrStr( ' z = ' ); WrHex( z ); WrLn();
        WrStr( ' r = ' ); WrHex( r ); WrLn();
        Fatal( 'Test41: Wrong result' );
      END;
    END;
  END;

  IO.WrStr( 'Test41: end ' ); IO.WrLngCard( Tmr0.Diff( t, Now() ), 1 ); IO.WrStr( 'ms' ); IO.WrLn();
END Test41;


BEGIN
  Test41();
END LrgCards.
